from typing import Optional
from funcnodes import Shelf, NodeDecorator
from exposedfunctionality import controlled_wrapper
from tensorflow.keras.optimizers import (
    Optimizer,
    SGD,
    RMSprop,
    Adam,
    AdamW,
    Adadelta,
    Adagrad,
    Adamax,
    Adafactor,
    Nadam,
    Ftrl,
    Lion,
)


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.SGD",
    name="SGD",
)
@controlled_wrapper(SGD, wrapper_attribute="__fnwrapped__")
def _SGD(
    learning_rate: float = 0.01,
    momentum: float = 0.0,
    nesterov: bool = False,
    weight_decay: Optional[float] = None,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return SGD(
        learning_rate=learning_rate,
        momentum=momentum,
        nesterov=nesterov,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.RMSprop",
    name="RMSprop",
)
@controlled_wrapper(RMSprop, wrapper_attribute="__fnwrapped__")
def _RMSprop(
    learning_rate: float = 0.001,
    rho: float = 0.9,
    momentum: float = 0.0,
    epsilon: float = 1e-07,
    centered: bool = False,
    weight_decay: Optional[float] = None,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return RMSprop(
        learning_rate=learning_rate,
        momentum=momentum,
        rho=rho,
        epsilon=epsilon,
        centered=centered,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.Adam",
    name="Adam",
)
@controlled_wrapper(Adam, wrapper_attribute="__fnwrapped__")
def _Adam(
    learning_rate: float = 0.001,
    beta_1: float = 0.9,
    beta_2: float = 0.999,
    epsilon: float = 1e-07,
    amsgrad: bool = False,
    weight_decay: Optional[float] = None,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return Adam(
        learning_rate=learning_rate,
        beta_1=beta_1,
        beta_2=beta_2,
        epsilon=epsilon,
        amsgrad=amsgrad,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.AdamW",
    name="AdamW",
)
@controlled_wrapper(AdamW, wrapper_attribute="__fnwrapped__")
def _AdamW(
    learning_rate: float = 0.001,
    weight_decay: float = 0.004,
    beta_1: float = 0.9,
    beta_2: float = 0.999,
    epsilon: float = 1e-07,
    amsgrad: bool = False,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return AdamW(
        learning_rate=learning_rate,
        beta_1=beta_1,
        beta_2=beta_2,
        epsilon=epsilon,
        amsgrad=amsgrad,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.Adadelta",
    name="Adadelta",
)
@controlled_wrapper(Adadelta, wrapper_attribute="__fnwrapped__")
def _Adadelta(
    learning_rate: float = 0.001,
    rho: float = 0.95,
    weight_decay: Optional[float] = None,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return Adadelta(
        learning_rate=learning_rate,
        rho=rho,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.Adagrad",
    name="Adagrad",
)
@controlled_wrapper(Adagrad, wrapper_attribute="__fnwrapped__")
def _Adagrad(
    learning_rate: float = 0.001,
    initial_accumulator_value: float = 0.1,
    epsilon: float = 1e-07,
    weight_decay: Optional[float] = None,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return Adagrad(
        learning_rate=learning_rate,
        initial_accumulator_value=initial_accumulator_value,
        epsilon=epsilon,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.Adamax",
    name="Adamax",
)
@controlled_wrapper(Adamax, wrapper_attribute="__fnwrapped__")
def _Adamax(
    learning_rate: float = 0.001,
    beta_1: float = 0.9,
    beta_2: float = 0.999,
    epsilon: float = 1e-07,
    weight_decay: Optional[float] = None,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return Adamax(
        learning_rate=learning_rate,
        beta_1=beta_1,
        beta_2=beta_2,
        epsilon=epsilon,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.Adafactor",
    name="Adafactor",
)
@controlled_wrapper(Adafactor, wrapper_attribute="__fnwrapped__")
def _Adafactor(
    learning_rate: float = 0.001,
    beta_2_decay: float = -0.8,
    epsilon_1: float = 1e-30,
    epsilon_2: float = 0.001,
    clip_threshold: float = 1.0,
    relative_step: bool = True,
    weight_decay: Optional[float] = None,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return Adafactor(
        learning_rate=learning_rate,
        beta_2_decay=beta_2_decay,
        epsilon_1=epsilon_1,
        epsilon_2=epsilon_2,
        clip_threshold=clip_threshold,
        relative_step=relative_step,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.Nadam",
    name="Nadam",
)
@controlled_wrapper(Nadam, wrapper_attribute="__fnwrapped__")
def _Nadam(
    learning_rate: float = 0.001,
    beta_1: float = 0.9,
    beta_2: float = 0.999,
    epsilon: float = 1e-07,
    weight_decay: Optional[float] = None,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return Nadam(
        learning_rate=learning_rate,
        beta_1=beta_1,
        beta_2=beta_2,
        epsilon=epsilon,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.Ftrl",
    name="Ftrl",
)
@controlled_wrapper(Ftrl, wrapper_attribute="__fnwrapped__")
def _Ftrl(
    learning_rate: float = 0.001,
    learning_rate_power: float = -0.5,
    initial_accumulator_value: float = 0.1,
    l1_regularization_strength: float = 0.0,
    l2_regularization_strength: float = 0.0,
    l2_shrinkage_regularization_strength: float = 0.0,
    beta: float = 0.0,
    weight_decay: Optional[float] = None,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return Ftrl(
        learning_rate=learning_rate,
        learning_rate_power=learning_rate_power,
        initial_accumulator_value=initial_accumulator_value,
        l1_regularization_strength=l1_regularization_strength,
        l2_regularization_strength=l2_regularization_strength,
        l2_shrinkage_regularization_strength=l2_shrinkage_regularization_strength,
        beta=beta,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


@NodeDecorator(
    node_id="tensorflow.keras.optimizers.Lion",
    name="Lion",
)
@controlled_wrapper(Lion, wrapper_attribute="__fnwrapped__")
def _Lion(
    learning_rate: float = 0.001,
    beta_1: float = 0.9,
    beta_2: float = 0.999,
    weight_decay: Optional[float] = None,
    clipnorm: Optional[float] = None,
    clipvalue: Optional[float] = None,
    global_clipnorm: Optional[float] = None,
    use_ema: bool = False,
    ema_momentum: float = 0.99,
    ema_overwrite_frequency: Optional[int] = None,
    loss_scale_factor: Optional[float] = None,
    gradient_accumulation_steps: Optional[int] = None,
) -> Optimizer:
    return Lion(
        learning_rate=learning_rate,
        beta_1=beta_1,
        beta_2=beta_2,
        weight_decay=weight_decay,
        clipnorm=clipnorm,
        clipvalue=clipvalue,
        global_clipnorm=global_clipnorm,
        use_ema=use_ema,
        ema_momentum=ema_momentum,
        ema_overwrite_frequency=ema_overwrite_frequency,
        loss_scale_factor=loss_scale_factor,
        gradient_accumulation_steps=gradient_accumulation_steps,
    )


OPTIMIZERS_NODE_SHELFE = Shelf(
    nodes=[
        _SGD,
        _RMSprop,
        _Adagrad,
        _Adadelta,
        _Adam,
        _Adamax,
        _Nadam,
        _Ftrl,
        _Lion,
    ],
    subshelves=[],
    name="Optimizers",
    description="An optimizer is one of the two arguments required for compiling a Keras model:",
)



from typing import Union, Optional, Iterator, Tuple, Callable
import numpy as np
from funcnodes import Shelf, NodeDecorator
from exposedfunctionality import controlled_wrapper
from enum import Enum
from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model
from tensorflow.keras.applications import (
    Xception,
    VGG16,
    VGG19,
    ResNet50,
    ResNet50V2,
    ResNet101,
    ResNet101V2,
    ResNet152,
    ResNet152V2,
    InceptionV3,
    InceptionResNetV2,
    MobileNet,
    MobileNetV2,
    MobileNetV3Small,
    MobileNetV3Large,
    DenseNet121,
    DenseNet169,
    DenseNet201,
    NASNetLarge,
    NASNetMobile,
    EfficientNetB0,
    EfficientNetB1,
    EfficientNetB2,
    EfficientNetB3,
    EfficientNetB4,
    EfficientNetB5,
    EfficientNetB6,
    EfficientNetB7,
    EfficientNetV2B0,
    EfficientNetV2B1,
    EfficientNetV2B2,
    EfficientNetV2B3,
    EfficientNetV2S,
    EfficientNetV2M,
    EfficientNetV2L,
    ConvNeXtTiny,
    ConvNeXtSmall,
    ConvNeXtBase,
    ConvNeXtLarge,
    ConvNeXtXLarge,
)


class Weights(Enum):
    imagenet = "imagenet"
    NONE = None

    @classmethod
    def default(cls):
        return cls.imagenet.value


class Pooling(Enum):
    avg = "avg"
    max = "max"
    NONE = None

    @classmethod
    def default(cls):
        return cls.NONE.value


class ClassifierActivation(Enum):
    softmax = "softmax"
    relu = "relu"
    sigmoid = "sigmoid"
    tanh = "tanh"
    linear = "linear"
    selu = "selu"
    elu = "elu"
    exponential = "exponential"
    swish = "swish"
    gelu = "gelu"
    softplus = "softplus"
    softsign = "softsign"

    @classmethod
    def default(cls):
        return cls.softmax.value


@NodeDecorator(
    node_id="tensorflow.keras.applications.Xception",
    name="Xception",
)
@controlled_wrapper(Xception, wrapper_attribute="__fnwrapped__")
def _Xception(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 299,
    input_shape_width: int = 299,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (299, 299, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (229, 229)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 71 or input_shape[1] < 71:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return Xception(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.VGG16",
    name="VGG16",
)
@controlled_wrapper(VGG16, wrapper_attribute="__fnwrapped__")
def _VGG16(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 31 or input_shape[1] < 31:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 31."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return VGG16(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.VGG19",
    name="VGG19",
)
@controlled_wrapper(VGG19, wrapper_attribute="__fnwrapped__")
def _VGG19(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 31 or input_shape[1] < 31:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 31."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return VGG19(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ResNet50",
    name="ResNet50",
)
@controlled_wrapper(ResNet50, wrapper_attribute="__fnwrapped__")
def _ResNet50(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 31 or input_shape[1] < 31:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 31."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ResNet50(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ResNet50V2",
    name="ResNet50V2",
)
@controlled_wrapper(ResNet50V2, wrapper_attribute="__fnwrapped__")
def _ResNet50V2(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 31 or input_shape[1] < 31:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 31."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ResNet50V2(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ResNet101",
    name="ResNet101",
)
@controlled_wrapper(ResNet101, wrapper_attribute="__fnwrapped__")
def _ResNet101(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 31 or input_shape[1] < 31:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 31."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ResNet101(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ResNet101V2",
    name="ResNet101V2",
)
@controlled_wrapper(ResNet101V2, wrapper_attribute="__fnwrapped__")
def _ResNet101V2(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 31 or input_shape[1] < 31:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 31."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ResNet101V2(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ResNet152",
    name="ResNet152",
)
@controlled_wrapper(ResNet152, wrapper_attribute="__fnwrapped__")
def _ResNet152(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 31 or input_shape[1] < 31:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 31."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ResNet152(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ResNet152V2",
    name="ResNet152V2",
)
@controlled_wrapper(ResNet152V2, wrapper_attribute="__fnwrapped__")
def _ResNet152V2(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 31 or input_shape[1] < 31:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 31."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ResNet152V2(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.InceptionV3",
    name="InceptionV3",
)
@controlled_wrapper(InceptionV3, wrapper_attribute="__fnwrapped__")
def _InceptionV3(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 299,
    input_shape_width: int = 299,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (299, 299, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (299, 299)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 31 or input_shape[1] < 31:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 31."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return InceptionV3(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.InceptionResNetV2",
    name="InceptionResNetV2",
)
@controlled_wrapper(InceptionResNetV2, wrapper_attribute="__fnwrapped__")
def _InceptionResNetV2(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 299,
    input_shape_width: int = 299,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (299, 299, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (299, 299)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 31 or input_shape[1] < 31:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 31."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return InceptionResNetV2(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.MobileNet",
    name="MobileNet",
)
@controlled_wrapper(MobileNet, wrapper_attribute="__fnwrapped__")
def _MobileNet(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    alpha: float = 1.0,
    depth_multiplier: int = 1,
    dropout: float = 0.001,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 32."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return MobileNet(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        alpha=alpha,
        depth_multiplier=depth_multiplier,
        dropout=dropout,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.MobileNetV2",
    name="MobileNetV2",
)
@controlled_wrapper(MobileNetV2, wrapper_attribute="__fnwrapped__")
def _MobileNetV2(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    alpha: float = 1.0,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 32."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return MobileNetV2(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        alpha=alpha,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.MobileNetV3Small",
    name="MobileNetV3Small",
)
@controlled_wrapper(MobileNetV3Small, wrapper_attribute="__fnwrapped__")
def _MobileNetV3Small(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    alpha: float = 1.0,
    minimalistic: bool = False,
    dropout_rate: float = 0.2,
    include_preprocessing: bool = True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 32."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return MobileNetV3Small(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        alpha=alpha,
        minimalistic=minimalistic,
        dropout_rate=dropout_rate,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.MobileNetV3Large",
    name="MobileNetV3Large",
)
@controlled_wrapper(MobileNetV3Large, wrapper_attribute="__fnwrapped__")
def _MobileNetV3Large(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    alpha: float = 1.0,
    minimalistic: bool = False,
    dropout_rate: float = 0.2,
    include_preprocessing: bool = True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 32."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return MobileNetV3Large(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        alpha=alpha,
        minimalistic=minimalistic,
        dropout_rate=dropout_rate,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.DenseNet121",
    name="DenseNet121",
)
@controlled_wrapper(DenseNet121, wrapper_attribute="__fnwrapped__")
def _DenseNet121(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return DenseNet121(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.DenseNet169",
    name="DenseNet169",
)
@controlled_wrapper(DenseNet169, wrapper_attribute="__fnwrapped__")
def _DenseNet169(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return DenseNet169(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.DenseNet201",
    name="DenseNet201",
)
@controlled_wrapper(DenseNet201, wrapper_attribute="__fnwrapped__")
def _DenseNet201(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return DenseNet201(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.NASNetLarge",
    name="NASNetLarge",
)
@controlled_wrapper(NASNetLarge, wrapper_attribute="__fnwrapped__")
def _NASNetLarge(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 331,
    input_shape_width: int = 331,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (331, 331, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (331, 331)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return NASNetLarge(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.NASNetMobile",
    name="NASNetMobile",
)
@controlled_wrapper(NASNetMobile, wrapper_attribute="__fnwrapped__")
def _NASNetMobile(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return NASNetMobile(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetB0",
    name="EfficientNetB0",
)
@controlled_wrapper(EfficientNetB0, wrapper_attribute="__fnwrapped__")
def _EfficientNetB0(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetB0(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetB1",
    name="EfficientNetB1",
)
@controlled_wrapper(EfficientNetB1, wrapper_attribute="__fnwrapped__")
def _EfficientNetB1(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 240,
    input_shape_width: int = 240,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (240, 240, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (240, 240)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetB1(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetB2",
    name="EfficientNetB2",
)
@controlled_wrapper(EfficientNetB2, wrapper_attribute="__fnwrapped__")
def _EfficientNetB2(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 260,
    input_shape_width: int = 260,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (260, 260, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (260, 260)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetB2(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetB3",
    name="EfficientNetB3",
)
@controlled_wrapper(EfficientNetB1, wrapper_attribute="__fnwrapped__")
def _EfficientNetB3(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 300,
    input_shape_width: int = 300,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (300, 300, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (300, 300)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetB3(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetB4",
    name="EfficientNetB4",
)
@controlled_wrapper(EfficientNetB1, wrapper_attribute="__fnwrapped__")
def _EfficientNetB4(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 380,
    input_shape_width: int = 380,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (380, 380, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (380, 380)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetB4(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetB5",
    name="EfficientNetB5",
)
@controlled_wrapper(EfficientNetB1, wrapper_attribute="__fnwrapped__")
def _EfficientNetB5(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 456,
    input_shape_width: int = 456,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (456, 456, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (456, 456)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetB5(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetB6",
    name="EfficientNetB6",
)
@controlled_wrapper(EfficientNetB1, wrapper_attribute="__fnwrapped__")
def _EfficientNetB6(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 528,
    input_shape_width: int = 528,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (528, 528, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (528, 528)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetB6(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetB7",
    name="EfficientNetB7",
)
@controlled_wrapper(EfficientNetB1, wrapper_attribute="__fnwrapped__")
def _EfficientNetB7(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 600,
    input_shape_width: int = 600,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (600, 600, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (600, 600)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetB7(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetV2B0",
    name="EfficientNetV2B0",
)
@controlled_wrapper(EfficientNetB1, wrapper_attribute="__fnwrapped__")
def _EfficientNetV2B0(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 380,
    input_shape_width: int = 380,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (380, 380, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (380, 380)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetV2B0(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetV2B1",
    name="EfficientNetV2B1",
)
@controlled_wrapper(EfficientNetV2B1, wrapper_attribute="__fnwrapped__")
def _EfficientNetV2B1(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 240,
    input_shape_width: int = 240,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (240, 240, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (240, 240)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetV2B1(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetV2B2",
    name="EfficientNetV2B2",
)
@controlled_wrapper(EfficientNetV2B2, wrapper_attribute="__fnwrapped__")
def _EfficientNetV2B2(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 260,
    input_shape_width: int = 260,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (260, 260, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (260, 260)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetV2B2(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetV2B3",
    name="EfficientNetV2B3",
)
@controlled_wrapper(EfficientNetV2B3, wrapper_attribute="__fnwrapped__")
def _EfficientNetV2B3(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 300,
    input_shape_width: int = 300,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (300, 300, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (300, 300)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetV2B3(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetV2S",
    name="EfficientNetV2S",
)
@controlled_wrapper(EfficientNetB1, wrapper_attribute="__fnwrapped__")
def _EfficientNetV2S(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 384,
    input_shape_width: int = 384,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (384, 384, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (384, 384)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetV2S(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetV2M",
    name="EfficientNetV2M",
)
@controlled_wrapper(EfficientNetV2M, wrapper_attribute="__fnwrapped__")
def _EfficientNetV2M(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 480,
    input_shape_width: int = 480,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (480, 480, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (480, 480)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetV2M(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.EfficientNetV2L",
    name="EfficientNetV2L",
)
@controlled_wrapper(EfficientNetV2L, wrapper_attribute="__fnwrapped__")
def _EfficientNetV2L(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 480,
    input_shape_width: int = 480,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (480, 480, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (480, 480)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return EfficientNetV2L(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ConvNeXtTiny",
    name="ConvNeXtTiny",
)
@controlled_wrapper(ConvNeXtTiny, wrapper_attribute="__fnwrapped__")
def _ConvNeXtTiny(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ConvNeXtTiny(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ConvNeXtSmall",
    name="ConvNeXtSmall",
)
@controlled_wrapper(ConvNeXtSmall, wrapper_attribute="__fnwrapped__")
def _ConvNeXtSmall(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ConvNeXtSmall(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ConvNeXtBase",
    name="ConvNeXtBase",
)
@controlled_wrapper(ConvNeXtBase, wrapper_attribute="__fnwrapped__")
def _ConvNeXtBase(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ConvNeXtBase(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ConvNeXtLarge",
    name="ConvNeXtLarge",
)
@controlled_wrapper(ConvNeXtLarge, wrapper_attribute="__fnwrapped__")
def _ConvNeXtLarge(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ConvNeXtLarge(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


@NodeDecorator(
    node_id="tensorflow.keras.applications.ConvNeXtXLarge",
    name="ConvNeXtXLarge",
)
@controlled_wrapper(ConvNeXtXLarge, wrapper_attribute="__fnwrapped__")
def _ConvNeXtXLarge(
    include_top: bool = True,
    weights: Weights = Weights.default(),
    input_tensor: Optional[Input] = None,
    input_shape_height: int = 224,
    input_shape_width: int = 224,
    pooling: Pooling = Pooling.default(),
    classes: int = 1000,
    classifier_activation: Optional[
        Union[ClassifierActivation, Callable]
    ] = ClassifierActivation.default(),
    include_preprocessing=True,
) -> Callable[[], Model]:
    if isinstance(classifier_activation, ClassifierActivation):
        classifier_activation = classifier_activation.value
    if isinstance(pooling, Pooling):
        pooling = pooling.value
    if isinstance(weights, Weights):
        weights = weights.value

    input_shape: Tuple[int, int, int] = (input_shape_height, input_shape_width, 3)
    # Validate input_shape
    if input_shape is not None:
        if include_top and input_shape != (224, 224, 3):
            raise ValueError(
                "if include_top is True, the (input_shape_height, input_shape_width) has to be (224, 224)."
            )
        if len(input_shape) != 3 or input_shape[2] != 3:
            raise ValueError(
                "input_shape must be a tuple of three integers (height, width, 3)."
            )
        if input_shape[0] < 32 or input_shape[1] < 32:
            raise ValueError(
                "input_shape dimensions (height, width) must be no smaller than 71."
            )
    elif not include_top:
        input_shape = (input_shape_height, input_shape_width, 3)

    return ConvNeXtXLarge(
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        classes=classes,
        classifier_activation=classifier_activation,
        include_preprocessing=include_preprocessing,
    )


APPLICATION_NODE_SHELFE = Shelf(
    nodes=[
        _Xception,
        _VGG16,
        _VGG19,
        _ResNet50,
        _ResNet50V2,
        _ResNet101,
        _ResNet101V2,
        _ResNet152,
        _ResNet152V2,
        _InceptionV3,
        _InceptionResNetV2,
        _MobileNet,
        _MobileNetV2,
        _MobileNetV3Small,
        _MobileNetV3Large,
        _DenseNet121,
        _DenseNet169,
        _DenseNet201,
        _NASNetLarge,
        _NASNetMobile,
        _EfficientNetB0,
        _EfficientNetB1,
        _EfficientNetB1,
        _EfficientNetB2,
        _EfficientNetB3,
        _EfficientNetB4,
        _EfficientNetB5,
        _EfficientNetB6,
        _EfficientNetB7,
        _EfficientNetV2B0,
        _EfficientNetV2B1,
        _EfficientNetV2B2,
        _EfficientNetV2B2,
        _EfficientNetV2B3,
        _EfficientNetV2S,
        _EfficientNetV2M,
        _EfficientNetV2L,
        _ConvNeXtTiny,
        _ConvNeXtSmall,
        _ConvNeXtBase,
        _ConvNeXtLarge,
        _ConvNeXtXLarge,
    ],
    subshelves=[],
    name="Applications",
    description="Keras Applications are deep learning models that are made available alongside pre-trained weights. These models can be used for prediction, feature extraction, and fine-tuning.Weights are downloaded automatically when instantiating a model. They are stored at ~/.keras/models/.Upon instantiation, the models will be built according to the image data format set in your Keras configuration file at ~/.keras/keras.json. For instance, if you have set image_data_format=channels_last, then any model loaded from this repository will get built according to the data format convention 'Height-Width-Depth'.",
)



import funcnodes as fn
from .applications import APPLICATION_NODE_SHELFE
from .fit import FIT_NODE_SHELFE
from .datasets import DATASETS_NODE_SHELFE
from .metrics import METRICS_NODE_SHELFE
from .optimizers import OPTIMIZERS_NODE_SHELFE
from .losses import LOSSES_NODE_SHELFE
from .layers import LAYERS_NODE_SHELFE
from .models import MODELS_NODE_SHELFE
from .utilities import UTILS_NODE_SHELFE

__version__ = "0.1.7"

NODE_SHELF = fn.Shelf(
    name="Keras",
    description="Tensorflow-Keras for funcnodes",
    nodes=[],
    subshelves=[
        LAYERS_NODE_SHELFE,
        MODELS_NODE_SHELFE,
        FIT_NODE_SHELFE,
        OPTIMIZERS_NODE_SHELFE,
        METRICS_NODE_SHELFE,
        LOSSES_NODE_SHELFE,
        UTILS_NODE_SHELFE,
        DATASETS_NODE_SHELFE,
        APPLICATION_NODE_SHELFE,
    ],
)



from typing import Literal, Union, Optional, Iterator, Tuple, Callable, List
import numpy as np
from funcnodes import Shelf, NodeDecorator
from exposedfunctionality import controlled_wrapper
from enum import Enum
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Optimizer
from tensorflow.keras.metrics import Metric
from tensorflow.keras.callbacks import Callback, History
from tensorflow.keras.datasets import (
    mnist,
    cifar10,
    cifar100,
    imdb,
    reuters,
    fashion_mnist,
    california_housing,
)


@NodeDecorator(
    node_id="tensorflow.keras.datasets.mnist ",
    name="mnist ",
    outputs=[
        {"name": "x_train"},
        {"name": "y_train"},
        {"name": "x_test"},
        {"name": "y_test"},
    ],
)
def _mnist() -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    (x_train, y_train), (x_test, y_test) = mnist.load_data()
    return x_train, y_train, x_test, y_test


@NodeDecorator(
    node_id="tensorflow.keras.datasets.cifar10 ",
    name="cifar10 ",
    outputs=[
        {"name": "x_train"},
        {"name": "y_train"},
        {"name": "x_test"},
        {"name": "y_test"},
    ],
)
def _cifar10() -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    (x_train, y_train), (x_test, y_test) = cifar10.load_data()
    return x_train, y_train, x_test, y_test


class LabelMode(Enum):
    fine = "fine"
    coarse = "coarse"

    @classmethod
    def default(cls):
        return cls.fine.value


@NodeDecorator(
    node_id="tensorflow.keras.datasets.cifar100 ",
    name="cifar100 ",
    outputs=[
        {"name": "x_train"},
        {"name": "y_train"},
        {"name": "x_test"},
        {"name": "y_test"},
    ],
)
def _cifar100(
    label_mode: LabelMode = LabelMode.default(),
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    if isinstance(label_mode, LabelMode):
        label_mode = label_mode.value
    (x_train, y_train), (x_test, y_test) = cifar100.load_data(label_mode=label_mode)
    return x_train, y_train, x_test, y_test


@NodeDecorator(
    node_id="tensorflow.keras.datasets.imdb ",
    name="imdb ",
    outputs=[
        {"name": "x_train"},
        {"name": "y_train"},
        {"name": "x_test"},
        {"name": "y_test"},
    ],
)
def _imdb(
    num_words: Optional[int] = None,
    skip_top: int = 0,
    maxlen: Optional[int] = None,
    seed: int = 113,
    start_char: int = 1,
    oov_char: int = 2,
    index_from: int = 3,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    (x_train, y_train), (x_test, y_test) = imdb.load_data(
        num_words=num_words,
        skip_top=skip_top,
        maxlen=maxlen,
        seed=seed,
        start_char=start_char,
        oov_char=oov_char,
        index_from=index_from,
    )
    return x_train, y_train, x_test, y_test


@NodeDecorator(
    node_id="tensorflow.keras.datasets.reuters ",
    name="reuters ",
    outputs=[
        {"name": "x_train"},
        {"name": "y_train"},
        {"name": "x_test"},
        {"name": "y_test"},
    ],
)
def _reuters(
    num_words: Optional[int] = None,
    skip_top: int = 0,
    maxlen: Optional[int] = None,
    test_split: float = 0.2,
    seed: int = 113,
    start_char: int = 1,
    oov_char: int = 2,
    index_from: int = 3,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    if not 0 <= test_split < 1:
        raise ValueError("test_split must be between 0 and 1")
    (x_train, y_train), (x_test, y_test) = reuters.load_data(
        num_words=num_words,
        skip_top=skip_top,
        maxlen=maxlen,
        test_split=test_split,
        seed=seed,
        start_char=start_char,
        oov_char=oov_char,
        index_from=index_from,
    )
    return x_train, y_train, x_test, y_test


@NodeDecorator(
    node_id="tensorflow.keras.datasets.fashion_mnist ",
    name="fashion_mnist ",
    outputs=[
        {"name": "x_train"},
        {"name": "y_train"},
        {"name": "x_test"},
        {"name": "y_test"},
    ],
)
def _fashion_mnist() -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    (x_train, y_train), (x_test, y_test) = fashion_mnist.load_data()
    return x_train, y_train, x_test, y_test


class VersionMode(Enum):
    small = "small"
    large = "large"

    @classmethod
    def default(cls):
        return cls.small.value


@NodeDecorator(
    node_id="tensorflow.keras.datasets.california_housing ",
    name="california_housing ",
    outputs=[
        {"name": "x_train"},
        {"name": "y_train"},
        {"name": "x_test"},
        {"name": "y_test"},
    ],
)
def _california_housing(
    version: VersionMode = VersionMode.default(),
    test_split: float = 0.2,
    seed: int = 113,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    if isinstance(version, VersionMode):
        version = version.value
    if not 0 <= test_split < 1:
        raise ValueError("test_split must be between 0 and 1")
    (x_train, y_train), (x_test, y_test) = california_housing.load_data(
        version=version, test_split=test_split, seed=seed
    )
    return x_train, y_train, x_test, y_test


DATASETS_NODE_SHELFE = Shelf(
    nodes=[
        _mnist,
        _cifar10,
        _cifar100,
        _imdb,
        _reuters,
        _fashion_mnist,
        _california_housing,
    ],
    subshelves=[],
    name="Datasets",
    description="The keras.datasets module provide a few toy datasets (already-vectorized, in Numpy format) that can be used for debugging a model or creating simple code examples.",
)



from typing import Optional, Union, Sequence, Callable
from enum import Enum
from funcnodes import Shelf, NodeDecorator
from exposedfunctionality import controlled_wrapper
from tensorflow.keras import KerasTensor
from tensorflow.keras.initializers import Initializer
from tensorflow.keras.regularizers import Regularizer
from tensorflow.keras.constraints import Constraint
from tensorflow.keras.layers import (
    Input,
    Dense,
    MaxPooling1D,
    MaxPooling2D,
    MaxPooling3D,
    AveragePooling1D,
    AveragePooling2D,
    AveragePooling3D,
    GlobalMaxPooling1D,
    GlobalMaxPooling2D,
    GlobalMaxPooling3D,
    GlobalAveragePooling1D,
    GlobalAveragePooling2D,
    GlobalAveragePooling3D,
    Conv1D,
    Conv2D,
    Conv3D,
    Dropout,
    Flatten,
)


class DataType(Enum):
    float16 = "float16"
    float32 = "float32"
    float64 = "float64"
    int16 = "int16"
    int32 = "int32"
    int64 = "int64"
    uint8 = "uint8"
    bool = "bool"
    string = "string"
    NONE = None

    @classmethod
    def default(cls):
        return cls.NONE.value


@NodeDecorator(
    node_id="tensorflow.keras.layers.Input",
    name="Input",
)
@controlled_wrapper(Input, wrapper_attribute="__fnwrapped__")
def _Input(
    shape_1_height_or_sequence_length: Optional[int] = None,
    shape_2_width_or_input_dim: Optional[int] = None,
    shape_3_channels: Optional[int] = None,
    batch_size: Optional[int] = None,
    name: Optional[str] = None,
    dtype: DataType = DataType.default(),
    sparse: bool = False,
    tensor: Optional[KerasTensor] = None,
) -> Callable[[], KerasTensor]:
    if shape_3_channels is None:
        shape = (shape_1_height_or_sequence_length, shape_2_width_or_input_dim)
    if shape_3_channels is None and shape_2_width_or_input_dim == 1:
        shape = (shape_1_height_or_sequence_length,)
    if shape_3_channels is not None:
        shape = (
            shape_1_height_or_sequence_length,
            shape_2_width_or_input_dim,
            shape_3_channels,
        )
    if isinstance(dtype, DataType):
        dtype = dtype.value
    return Input(
        shape=shape,
        batch_size=batch_size,
        name=name,
        dtype=dtype,
        sparse=sparse,
        tensor=tensor,
    )


class Activation(Enum):
    relu = "relu"
    sigmoid = "sigmoid"
    softmax = "softmax"
    tanh = "tanh"
    linear = "linear"
    NONE = None

    @classmethod
    def default(cls):
        return cls.NONE.value


@NodeDecorator(
    node_id="tensorflow.keras.layers.Dense",
    name="Dense",
)
@controlled_wrapper(Dense, wrapper_attribute="__fnwrapped__")
def _Dense(
    input_model: KerasTensor,
    units: int,
    activation: Activation = Activation.default(),
    use_bias: bool = True,
    kernel_initializer: Union[str, Initializer] = "glorot_uniform",
    bias_initializer: Union[str, Initializer] = "zeros",
    kernel_regularizer: Optional[Regularizer] = None,
    bias_regularizer: Optional[Regularizer] = None,
    activity_regularizer: Optional[Regularizer] = None,
    kernel_constraint: Optional[Constraint] = None,
    bias_constraint: Optional[Constraint] = None,
    lora_rank: Optional[int] = None,
) -> Callable[[], KerasTensor]:
    if isinstance(activation, Activation):
        activation = activation.value
    return Dense(
        units=units,
        activation=activation,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        bias_regularizer=bias_regularizer,
        activity_regularizer=activity_regularizer,
        kernel_constraint=kernel_constraint,
        bias_constraint=bias_constraint,
        lora_rank=lora_rank,
    )(input_model)


CORELAYERS_NODE_SHELFE = Shelf(
    nodes=[_Input, _Dense],
    subshelves=[],
    name="Core",
    description="",
)


class Padding(Enum):
    valid = "valid"
    same = "same"

    @classmethod
    def default(cls):
        return cls.valid.value


@NodeDecorator(
    node_id="tensorflow.keras.layers.MaxPooling1D",
    name="MaxPooling1D",
)
@controlled_wrapper(MaxPooling1D, wrapper_attribute="__fnwrapped__")
def _MaxPooling1D(
    input_model: KerasTensor,
    pool_size: int = 2,
    strides: Optional[int] = None,
    padding: Padding = Padding.default(),
    data_format: DataType = DataType.default(),
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    if isinstance(padding, Padding):
        padding = padding.value
    return MaxPooling1D(
        pool_size=pool_size,
        strides=strides,
        padding=padding,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.MaxPooling2D",
    name="MaxPooling2D",
)
@controlled_wrapper(MaxPooling2D, wrapper_attribute="__fnwrapped__")
def _MaxPooling2D(
    input_model: KerasTensor,
    pool_size_1: int = 2,
    pool_size_2: int = 2,
    strides: Optional[int] = None,
    padding: Padding = Padding.default(),
    data_format: DataType = DataType.default(),
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    if isinstance(padding, Padding):
        padding = padding.value
    return MaxPooling2D(
        pool_size=(pool_size_1, pool_size_2),
        strides=strides,
        padding=padding,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.MaxPooling3D",
    name="MaxPooling3D",
)
@controlled_wrapper(MaxPooling3D, wrapper_attribute="__fnwrapped__")
def _MaxPooling3D(
    input_model: KerasTensor,
    pool_size_1: int = 2,
    pool_size_2: int = 2,
    pool_size_3: int = 2,
    strides: Optional[int] = None,
    padding: Padding = Padding.default(),
    data_format: DataType = DataType.default(),
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    if isinstance(padding, Padding):
        padding = padding.value
    return MaxPooling3D(
        pool_size=(pool_size_1, pool_size_2, pool_size_3),
        strides=strides,
        padding=padding,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.AveragePooling1D",
    name="AveragePooling1D",
)
@controlled_wrapper(AveragePooling1D, wrapper_attribute="__fnwrapped__")
def _AveragePooling1D(
    input_model: KerasTensor,
    pool_size: int = 2,
    strides: Optional[int] = None,
    padding: Padding = Padding.default(),
    data_format: DataType = DataType.default(),
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    if isinstance(padding, Padding):
        padding = padding.value
    return AveragePooling1D(
        pool_size=pool_size,
        strides=strides,
        padding=padding,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.AveragePooling2D",
    name="AveragePooling2D",
)
@controlled_wrapper(AveragePooling2D, wrapper_attribute="__fnwrapped__")
def _AveragePooling2D(
    input_model: KerasTensor,
    pool_size_1: int = 2,
    pool_size_2: int = 2,
    strides: Optional[int] = None,
    padding: Padding = Padding.default(),
    data_format: DataType = DataType.default(),
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    if isinstance(padding, Padding):
        padding = padding.value
    return AveragePooling2D(
        pool_size=(pool_size_1, pool_size_2),
        strides=strides,
        padding=padding,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.AveragePooling3D",
    name="AveragePooling3D",
)
@controlled_wrapper(AveragePooling3D, wrapper_attribute="__fnwrapped__")
def _AveragePooling3D(
    input_model: KerasTensor,
    pool_size_1: int = 2,
    pool_size_2: int = 2,
    pool_size_3: int = 2,
    strides: Optional[int] = None,
    padding: Padding = Padding.default(),
    data_format: DataType = DataType.default(),
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    if isinstance(padding, Padding):
        padding = padding.value
    return AveragePooling3D(
        pool_size=(pool_size_1, pool_size_2, pool_size_3),
        strides=strides,
        padding=padding,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.GlobalMaxPooling1D",
    name="GlobalMaxPooling1D",
)
@controlled_wrapper(GlobalMaxPooling1D, wrapper_attribute="__fnwrapped__")
def _GlobalMaxPooling1D(
    input_model: KerasTensor,
    data_format: DataType = DataType.default(),
    keepdims: bool = False,
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    return GlobalMaxPooling1D(
        keepdims=keepdims,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.GlobalMaxPooling2D",
    name="GlobalMaxPooling2D",
)
@controlled_wrapper(GlobalMaxPooling2D, wrapper_attribute="__fnwrapped__")
def _GlobalMaxPooling2D(
    input_model: KerasTensor,
    data_format: DataType = DataType.default(),
    keepdims: bool = False,
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    return GlobalMaxPooling2D(
        keepdims=keepdims,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.GlobalMaxPooling3D",
    name="GlobalMaxPooling3D",
)
@controlled_wrapper(GlobalMaxPooling3D, wrapper_attribute="__fnwrapped__")
def _GlobalMaxPooling3D(
    input_model: KerasTensor,
    data_format: DataType = DataType.default(),
    keepdims: bool = False,
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    return GlobalMaxPooling3D(
        keepdims=keepdims,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.GlobalAveragePooling1D",
    name="GlobalAveragePooling1D",
)
@controlled_wrapper(GlobalAveragePooling1D, wrapper_attribute="__fnwrapped__")
def _GlobalAveragePooling1D(
    input_model: KerasTensor,
    data_format: DataType = DataType.default(),
    keepdims: bool = False,
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    return GlobalAveragePooling1D(
        keepdims=keepdims,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.GlobalAveragePooling2D",
    name="GlobalAveragePooling2D",
)
@controlled_wrapper(GlobalAveragePooling2D, wrapper_attribute="__fnwrapped__")
def _GlobalAveragePooling2D(
    input_model: KerasTensor,
    data_format: DataType = DataType.default(),
    keepdims: bool = False,
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    return GlobalAveragePooling2D(
        keepdims=keepdims,
        data_format=data_format,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.GlobalAveragePooling3D",
    name="GlobalAveragePooling3D",
)
@controlled_wrapper(GlobalAveragePooling3D, wrapper_attribute="__fnwrapped__")
def _GlobalAveragePooling3D(
    input_model: KerasTensor,
    data_format: DataType = DataType.default(),
    keepdims: bool = False,
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    return GlobalAveragePooling3D(
        keepdims=keepdims,
        data_format=data_format,
    )(input_model)


POOLINGLAYERS_NODE_SHELFE = Shelf(
    nodes=[
        _MaxPooling1D,
        _MaxPooling2D,
        _MaxPooling3D,
        _AveragePooling1D,
        _AveragePooling2D,
        _AveragePooling3D,
        _GlobalMaxPooling1D,
        _GlobalMaxPooling2D,
        _GlobalMaxPooling3D,
        _GlobalAveragePooling1D,
        _GlobalAveragePooling2D,
        _GlobalAveragePooling3D,
    ],
    subshelves=[],
    name="Pooling",
    description="",
)


class ConvPadding(Enum):
    valid = "valid"
    same = "same"
    causal = "causal"

    @classmethod
    def default(cls):
        return cls.valid.value


@NodeDecorator(
    node_id="tensorflow.keras.layers.Conv1D",
    name="Conv1D",
)
@controlled_wrapper(Conv1D, wrapper_attribute="__fnwrapped__")
def _Conv1D(
    input_model: KerasTensor,
    filters: int,
    kernel_size: int,
    strides: Optional[int] = None,
    padding: ConvPadding = ConvPadding.default(),
    data_format: DataType = DataType.default(),
    dilation_rate: int = 1,
    groups: int = 1,
    activation: Activation = Activation.default(),
    use_bias: bool = True,
    kernel_initializer: Union[str, Initializer] = "glorot_uniform",
    bias_initializer: Union[str, Initializer] = "zeros",
    kernel_regularizer: Optional[Regularizer] = None,
    bias_regularizer: Optional[Regularizer] = None,
    activity_regularizer: Optional[Regularizer] = None,
    kernel_constraint: Optional[Constraint] = None,
    bias_constraint: Optional[Constraint] = None,
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    if isinstance(padding, ConvPadding):
        padding = padding.value
    if isinstance(activation, Activation):
        activation = activation.value
    return Conv1D(
        filters=filters,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        data_format=data_format,
        dilation_rate=dilation_rate,
        groups=groups,
        activation=activation,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        bias_regularizer=bias_regularizer,
        activity_regularizer=activity_regularizer,
        kernel_constraint=kernel_constraint,
        bias_constraint=bias_constraint,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.Conv2D",
    name="Conv2D",
)
@controlled_wrapper(Conv2D, wrapper_attribute="__fnwrapped__")
def _Conv2D(
    input_model: KerasTensor,
    filters: int,
    kernel_size_1: int,
    kernel_size_2: int,
    strides_1: int = 1,
    strides_2: int = 1,
    padding: ConvPadding = ConvPadding.default(),
    data_format: DataType = DataType.default(),
    dilation_rate_1: int = 1,
    dilation_rate_2: int = 1,
    groups: int = 1,
    activation: Activation = Activation.default(),
    use_bias: bool = True,
    kernel_initializer: Union[str, Initializer] = "glorot_uniform",
    bias_initializer: Union[str, Initializer] = "zeros",
    kernel_regularizer: Optional[Regularizer] = None,
    bias_regularizer: Optional[Regularizer] = None,
    activity_regularizer: Optional[Regularizer] = None,
    kernel_constraint: Optional[Constraint] = None,
    bias_constraint: Optional[Constraint] = None,
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    if isinstance(padding, ConvPadding):
        padding = padding.value
    if isinstance(activation, Activation):
        activation = activation.value
    return Conv2D(
        filters=filters,
        kernel_size=(kernel_size_1, kernel_size_2),
        strides=(strides_1, strides_2),
        padding=padding,
        data_format=data_format,
        dilation_rate=(dilation_rate_1, dilation_rate_2),
        groups=groups,
        activation=activation,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        bias_regularizer=bias_regularizer,
        activity_regularizer=activity_regularizer,
        kernel_constraint=kernel_constraint,
        bias_constraint=bias_constraint,
    )(input_model)


@NodeDecorator(
    node_id="tensorflow.keras.layers.Conv3D",
    name="Conv3D",
)
@controlled_wrapper(Conv3D, wrapper_attribute="__fnwrapped__")
def _Conv3D(
    input_model: KerasTensor,
    filters: int,
    kernel_size_1: int,
    kernel_size_2: int,
    kernel_size_3: int,
    strides_1: int = 1,
    strides_2: int = 1,
    strides_3: int = 1,
    padding: ConvPadding = ConvPadding.default(),
    data_format: DataType = DataType.default(),
    dilation_rate_1: int = 1,
    dilation_rate_2: int = 1,
    dilation_rate_3: int = 1,
    groups: int = 1,
    activation: Activation = Activation.default(),
    use_bias: bool = True,
    kernel_initializer: Union[str, Initializer] = "glorot_uniform",
    bias_initializer: Union[str, Initializer] = "zeros",
    kernel_regularizer: Optional[Regularizer] = None,
    bias_regularizer: Optional[Regularizer] = None,
    activity_regularizer: Optional[Regularizer] = None,
    kernel_constraint: Optional[Constraint] = None,
    bias_constraint: Optional[Constraint] = None,
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    if isinstance(padding, ConvPadding):
        padding = padding.value
    if isinstance(activation, Activation):
        activation = activation.value
    return Conv3D(
        filters=filters,
        kernel_size=(kernel_size_1, kernel_size_2, kernel_size_3),
        strides=(strides_1, strides_2, strides_3),
        padding=padding.value,
        data_format=data_format,
        dilation_rate=(dilation_rate_1, dilation_rate_2, dilation_rate_3),
        groups=groups,
        activation=activation,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        bias_regularizer=bias_regularizer,
        activity_regularizer=activity_regularizer,
        kernel_constraint=kernel_constraint,
        bias_constraint=bias_constraint,
    )(input_model)


CONVLAYERS_NODE_SHELFE = Shelf(
    nodes=[_Conv1D, _Conv2D, _Conv3D],
    subshelves=[],
    name="Convolution",
    description="",
)


@NodeDecorator(
    node_id="tensorflow.keras.layers.Dropout",
    name="Dropout",
)
@controlled_wrapper(Dropout, wrapper_attribute="__fnwrapped__")
def _Dropout(
    input_model: KerasTensor,
    rate: float,
    noise_shape: Optional[Sequence[int]] = None,
    seed: Optional[int] = None,
) -> Callable[[], KerasTensor]:
    if not 0 <= rate < 1:
        raise ValueError("rate must be between 0 and 1")

    return Dropout(rate=rate, noise_shape=noise_shape, seed=seed)(input_model)


REGLAYERS_NODE_SHELFE = Shelf(
    nodes=[_Dropout],
    subshelves=[],
    name="Regularization",
    description="",
)


@NodeDecorator(
    node_id="tensorflow.keras.layers.Flatten",
    name="Flatten",
)
@controlled_wrapper(Flatten, wrapper_attribute="__fnwrapped__")
def _Flatten(
    input_model: KerasTensor,
    data_format: DataType = DataType.default(),
) -> Callable[[], KerasTensor]:
    if isinstance(data_format, DataType):
        data_format = data_format.value
    return Flatten(data_format=data_format)(input_model)


RESHAPINGLAYERS_NODE_SHELFE = Shelf(
    nodes=[_Flatten],
    subshelves=[],
    name="Reshaping",
    description="",
)


LAYERS_NODE_SHELFE = Shelf(
    nodes=[],
    subshelves=[
        CORELAYERS_NODE_SHELFE,
        POOLINGLAYERS_NODE_SHELFE,
        CONVLAYERS_NODE_SHELFE,
        REGLAYERS_NODE_SHELFE,
        RESHAPINGLAYERS_NODE_SHELFE,
    ],
    name="Layers",
    description="Layers are the basic building blocks of neural networks in Keras. A layer consists of a tensor-in tensor-out computation function (the layer's call method) and some state, held in TensorFlow variables (the layer's weights).",
)



from typing import Callable
from funcnodes import Shelf, NodeDecorator
from exposedfunctionality import controlled_wrapper
from tensorflow.keras.models import Model
from tensorflow.keras import KerasTensor, Sequential


@NodeDecorator(
    node_id="tensorflow.keras.models.Model",
    name="Model",
)
@controlled_wrapper(Model, wrapper_attribute="__fnwrapped__")
def _Model(
    input: KerasTensor,
    output: KerasTensor,
) -> Callable[[], Model]:
    return Model(inputs=input, outputs=output)


MODEL_NODE_SHELFE = Shelf(
    nodes=[_Model],
    subshelves=[],
    name="Models ",
    description="",
)


@NodeDecorator(
    node_id="tensorflow.keras.Sequential",
    name="Sequential",
)
@controlled_wrapper(Sequential, wrapper_attribute="__fnwrapped__")
def _Sequential() -> Callable[[], Model]:
    return Sequential()


@NodeDecorator(
    node_id="tensorflow.keras.add",
    name="add",
)
def _add(sequential_model: Model, input: KerasTensor) -> Callable[[], Model]:
    return sequential_model.add(input)


SEQUENTIAL_NODE_SHELFE = Shelf(
    nodes=[_Sequential, _add],
    subshelves=[],
    name="Sequential",
    description="",
)


MODELS_NODE_SHELFE = Shelf(
    nodes=[],
    subshelves=[MODEL_NODE_SHELFE],
    name="Models ",
    description="There are three ways to create Keras models:The Sequential model, which is very straightforward (a simple list of layers), but is limited to single-input, single-output stacks of layers (as the name gives away). The Functional API, which is an easy-to-use, fully-featured API that supports arbitrary model architectures. For most people and most use cases, this is what you should be using. This is the Keras 'industry strength' model. Model subclassing, where you implement everything from scratch on your own. Use this if you have complex, out-of-the-box research use cases.",
)



from typing import Optional, Union
from enum import Enum
from funcnodes import Shelf, NodeDecorator
from exposedfunctionality import controlled_wrapper
import os
from tensorflow.keras.callbacks import Callback, ModelCheckpoint


class Monitor(Enum):
    accuracy = "accuracy"
    loss = "loss"
    val_loss = "val_loss"
    val_accuracy = "val_accuracy"

    @classmethod
    def default(cls):
        return cls.val_loss.value


class Verbose(Enum):
    zero = 0
    one = 1

    @classmethod
    def default(cls):
        return cls.zero.value


class Mode(Enum):
    min = "min"
    max = "max"
    auto = "auto"

    @classmethod
    def default(cls):
        return cls.auto.value


@NodeDecorator(
    node_id="tensorflow.keras.callbacks.ModelCheckpoint",
    name="ModelCheckpoint",
)
@controlled_wrapper(ModelCheckpoint, wrapper_attribute="__fnwrapped__")
def _ModelCheckpoint(
    monitor: Monitor = Monitor.default(),
    verbose: Verbose = Verbose.default(),
    save_best_only: bool = False,
    mode: Mode = Mode.default(),
    save_weights_only: bool = False,
    save_freq: Optional[int] = None,
    initial_value_threshold: Optional[float] = None,
) -> Callback:
    if isinstance(monitor, Monitor):
        monitor = monitor.value
    if isinstance(verbose, Verbose):
        verbose = verbose.value
    if isinstance(mode, Mode):
        mode = mode.value

    if save_freq is None:
        save_freq = "epoch"

    filepath: Union[str, os.PathLike] = "model.{epoch:02d}-{monitor:.2f}.h5"

    return ModelCheckpoint(
        filepath=filepath,
        monitor=monitor,
        verbose=verbose,
        save_best_only=save_best_only,
        mode=mode,
        save_weights_only=save_weights_only,
        save_freq=save_freq,
        initial_value_threshold=initial_value_threshold,
    )





from typing import Optional, Union
import numpy as np
from funcnodes import Shelf, NodeDecorator
from exposedfunctionality import controlled_wrapper
from tensorflow.keras.utils import to_categorical


@NodeDecorator(
    node_id="tensorflow.keras.utils.to_categorical",
    name="to_categorical",
)
@controlled_wrapper(to_categorical, wrapper_attribute="__fnwrapped__")
def _to_categorical(
    x: Union[list, np.ndarray],
    num_classes: Optional[int] = None,
) -> np.ndarray:
    return to_categorical(x, num_classes)


UTILS_NODE_SHELFE = Shelf(
    nodes=[_to_categorical],
    subshelves=[],
    name="uTILITIES",
    description="Python & NumPy utilities",
)



from typing import Literal, Union, Optional, Tuple, List
import numpy as np
from funcnodes import Shelf, NodeDecorator
from enum import Enum
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Optimizer
from tensorflow.keras.losses import Loss
from tensorflow.keras.metrics import Metric
from tensorflow.keras.callbacks import Callback, History


@NodeDecorator(
    node_id="tensorflow.keras.training.compile ",
    name="compile ",
    outputs=[
        {"name": "compiled_model"},
    ],
)
# @controlled_wrapper(compile , wrapper_attribute="__fnwrapped__")
def _compile(
    model: Model,
    optimizer: Optimizer,
    loss: Loss,
    metrics: Union[Metric, List[Metric], dict],
    loss_weights: Optional[Union[list, dict]] = None,
    weighted_metrics: Optional[list] = None,
    run_eagerly: bool = False,
    steps_per_execution: int = 1,
    jit_compile: Union[bool, Literal["auto"]] = "auto",
    auto_scale_loss: bool = True,
) -> Model:
    if isinstance(metrics, Metric):
        metrics = [metrics]
    model.compile(
        optimizer=optimizer,
        loss=loss,
        loss_weights=loss_weights,
        metrics=metrics,
        weighted_metrics=weighted_metrics,
        run_eagerly=run_eagerly,
        steps_per_execution=steps_per_execution,
        jit_compile=jit_compile,
        auto_scale_loss=auto_scale_loss,
    )
    return model


class Verbose(Enum):
    auto = "auto"
    zero = 0
    one = 1
    two = 2

    @classmethod
    def default(cls):
        return cls.auto.value


@NodeDecorator(
    node_id="tensorflow.keras.training.fit ",
    name="fit ",
    outputs=[
        {"name": "fitted_model"},
        {"name": "history"},
        {"name": "metrics_dictionary"},
    ],
)
def _fit(
    model: Model,
    x: Union[list, np.ndarray, dict],
    y: Optional[Union[list, np.ndarray, dict]] = None,
    batch_size: Optional[int] = None,
    epochs: int = 1,
    verbose: Verbose = Verbose.default(),
    callbacks: Optional[Union[Callback, List[Callback]]] = None,
    validation_split: float = 0.0,
    x_val: Optional[Union[list, np.ndarray, dict]] = None,
    y_val: Optional[Union[list, np.ndarray, dict]] = None,
    shuffle: bool = True,
    class_weight: Optional[dict] = None,
    sample_weight: Optional[np.ndarray] = None,
    initial_epoch: int = 0,
    steps_per_epoch: Optional[int] = None,
    validation_steps: Optional[int] = None,
    validation_batch_size: Optional[int] = None,
    validation_freq: int = 1,
) -> Tuple[Model, History, dict]:
    if isinstance(callbacks, Callback):
        callbacks = [callbacks]
    if x_val is not None and y_val is not None:
        validation_data: Tuple = [x_val, y_val]
    else:
        validation_data = None

    if not 0 <= validation_split < 1:
        raise ValueError("validation_split must be between 0 and 1")

    out = model.fit(
        x=x,
        y=y,
        batch_size=batch_size,
        epochs=epochs,
        verbose=verbose,
        callbacks=callbacks,
        validation_split=validation_split,
        validation_data=validation_data,
        shuffle=shuffle,
        class_weight=class_weight,
        sample_weight=sample_weight,
        initial_epoch=initial_epoch,
        steps_per_epoch=steps_per_epoch,
        validation_steps=validation_steps,
        validation_batch_size=validation_batch_size,
        validation_freq=validation_freq,
    )

    return model, out, out.history


@NodeDecorator(
    node_id="tensorflow.keras.training.evaluate ",
    name="evaluate ",
)
def _evaluate(
    model: Model,
    x: Union[list, np.ndarray, dict],
    y: Optional[Union[list, np.ndarray, dict]] = None,
    batch_size: Optional[int] = None,
    verbose: Verbose = Verbose.default(),
    sample_weight: Optional[np.ndarray] = None,
    steps: Optional[int] = None,
    callbacks: Optional[List[Callback]] = None,
) -> dict:
    return model.evaluate(
        x=x,
        y=y,
        batch_size=batch_size,
        verbose=verbose,
        sample_weight=sample_weight,
        steps=steps,
        callbacks=callbacks,
        return_dict=True,
    )


@NodeDecorator(
    node_id="tensorflow.keras.training.predict ",
    name="predict ",
)
def _predict(
    model: Model,
    x: Union[np.ndarray, List[np.ndarray]],
    batch_size: Optional[int] = None,
    verbose: Verbose = Verbose.default(),
    steps: Optional[int] = None,
    callbacks: Optional[List[Callback]] = None,
) -> np.ndarray:
    return model.predict(
        x=x,
        batch_size=batch_size,
        verbose=verbose,
        steps=steps,
        callbacks=callbacks,
    )


@NodeDecorator(
    node_id="tensorflow.keras.training.train_on_batch ",
    name="train_on_batch ",
)
def _train_on_batch(
    model: Model,
    x: np.ndarray,
    y: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    class_weight: Optional[dict] = None,
) -> dict:
    return model.train_on_batch(
        x=x,
        y=y,
        sample_weight=sample_weight,
        class_weight=class_weight,
        return_dict=True,
    )


@NodeDecorator(
    node_id="tensorflow.keras.training.test_on_batch ",
    name="test_on_batch ",
)
def _test_on_batch(
    model: Model,
    x: np.ndarray,
    y: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
) -> dict:
    return model.test_on_batch(
        x=x,
        y=y,
        sample_weight=sample_weight,
        return_dict=True,
    )


@NodeDecorator(
    node_id="tensorflow.keras.training.predict_on_batch ",
    name="predict_on_batch ",
)
def _predict_on_batch(
    model: Model,
    x: np.ndarray,
) -> dict:
    return model.predict_on_batch(
        x=x,
    )


FIT_NODE_SHELFE = Shelf(
    nodes=[
        _fit,
        _compile,
        _evaluate,
        _predict,
        _train_on_batch,
        _test_on_batch,
        _predict_on_batch,
    ],
    subshelves=[],
    name="Fit",
    description="Methods for compiling, fitting, and more.",
)



from typing import Union, Optional, List
import numpy as np
from funcnodes import Shelf, NodeDecorator
from exposedfunctionality import controlled_wrapper
from enum import Enum
from tensorflow.keras.losses import (
    Loss,
    BinaryCrossentropy,
    BinaryFocalCrossentropy,
    CategoricalCrossentropy,
    CategoricalFocalCrossentropy,
    SparseCategoricalCrossentropy,
    Poisson,
    KLDivergence,
    CTC,
    MeanSquaredError,
    MeanAbsoluteError,
    MeanAbsolutePercentageError,
    MeanSquaredLogarithmicError,
    CosineSimilarity,
    Huber,
    LogCosh,
    Hinge,
    SquaredHinge,
    CategoricalHinge,
)


class Reduction(Enum):
    sum = "sum"
    sum_over_batch_size = "sum_over_batch_size"
    NONE = None

    @classmethod
    def default(cls):
        return cls.sum_over_batch_size.value


@NodeDecorator(
    node_id="tensorflow.keras.losses.BinaryCrossentropy",
    name="BinaryCrossentropy",
)
@controlled_wrapper(BinaryCrossentropy, wrapper_attribute="__fnwrapped__")
def _BinaryCrossentropy(
    from_logits: bool = True,
    label_smoothing: float = 0.0,
    axis: int = -1,
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return BinaryCrossentropy(
        from_logits=from_logits,
        label_smoothing=label_smoothing,
        axis=axis,
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.BinaryFocalCrossentropy",
    name="BinaryFocalCrossentropy",
)
@controlled_wrapper(BinaryFocalCrossentropy, wrapper_attribute="__fnwrapped__")
def _BinaryFocalCrossentropy(
    apply_class_balancing: bool = False,
    alpha: float = 0.25,
    gamma: float = 2.0,
    from_logits: bool = True,
    label_smoothing: float = 0.0,
    axis: int = -1,
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return BinaryFocalCrossentropy(
        apply_class_balancing=apply_class_balancing,
        alpha=alpha,
        gamma=gamma,
        from_logits=from_logits,
        label_smoothing=label_smoothing,
        axis=axis,
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.CategoricalCrossentropy",
    name="CategoricalCrossentropy",
)
@controlled_wrapper(CategoricalCrossentropy, wrapper_attribute="__fnwrapped__")
def _CategoricalCrossentropy(
    from_logits: bool = True,
    label_smoothing: float = 0.0,
    axis: int = -1,
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return CategoricalCrossentropy(
        from_logits=from_logits,
        label_smoothing=label_smoothing,
        axis=axis,
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.CategoricalFocalCrossentropy",
    name="CategoricalFocalCrossentropy",
)
@controlled_wrapper(CategoricalFocalCrossentropy, wrapper_attribute="__fnwrapped__")
def _CategoricalFocalCrossentropy(
    alpha: float = 0.25,
    gamma: float = 2.0,
    from_logits: bool = True,
    label_smoothing: float = 0.0,
    axis: int = -1,
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return CategoricalFocalCrossentropy(
        alpha=alpha,
        gamma=gamma,
        from_logits=from_logits,
        label_smoothing=label_smoothing,
        axis=axis,
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.SparseCategoricalCrossentropy",
    name="SparseCategoricalCrossentropy",
)
@controlled_wrapper(SparseCategoricalCrossentropy, wrapper_attribute="__fnwrapped__")
def _SparseCategoricalCrossentropy(
    from_logits: bool = True,
    ignore_class: Optional[int] = None,
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return SparseCategoricalCrossentropy(
        from_logits=from_logits,
        ignore_class=ignore_class,
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.Poisson",
    name="Poisson",
)
@controlled_wrapper(Poisson, wrapper_attribute="__fnwrapped__")
def _Poisson(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return Poisson(
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.KLDivergence",
    name="KLDivergence",
)
@controlled_wrapper(KLDivergence, wrapper_attribute="__fnwrapped__")
def _KLDivergence(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return KLDivergence(
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.CTC",
    name="CTC",
)
@controlled_wrapper(CTC, wrapper_attribute="__fnwrapped__")
def _CTC(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return CTC(
        reduction=reduction,
    )


PROBABILISTIC_NODE_SHELFE = Shelf(
    nodes=[
        _BinaryCrossentropy,
        _BinaryFocalCrossentropy,
        _CategoricalCrossentropy,
        _CategoricalFocalCrossentropy,
        _SparseCategoricalCrossentropy,
        _Poisson,
        _KLDivergence,
        _CTC,
    ],
    subshelves=[],
    name="Probabilistic ",
    description="",
)


@NodeDecorator(
    node_id="tensorflow.keras.losses.MeanSquaredError",
    name="MeanSquaredError",
)
@controlled_wrapper(MeanSquaredError, wrapper_attribute="__fnwrapped__")
def _MeanSquaredError(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return MeanSquaredError(
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.MeanAbsoluteError",
    name="MeanAbsoluteError",
)
@controlled_wrapper(MeanAbsoluteError, wrapper_attribute="__fnwrapped__")
def _MeanAbsoluteError(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return MeanAbsoluteError(
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.MeanAbsolutePercentageError",
    name="MeanAbsolutePercentageError",
)
@controlled_wrapper(MeanAbsolutePercentageError, wrapper_attribute="__fnwrapped__")
def _MeanAbsolutePercentageError(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return MeanAbsolutePercentageError(
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.MeanSquaredLogarithmicError",
    name="MeanSquaredLogarithmicError",
)
@controlled_wrapper(MeanSquaredLogarithmicError, wrapper_attribute="__fnwrapped__")
def _MeanSquaredLogarithmicError(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return MeanSquaredLogarithmicError(
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.CosineSimilarity",
    name="CosineSimilarity",
)
@controlled_wrapper(CosineSimilarity, wrapper_attribute="__fnwrapped__")
def _CosineSimilarity(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return CosineSimilarity(
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.Huber",
    name="Huber",
)
@controlled_wrapper(Huber, wrapper_attribute="__fnwrapped__")
def _Huber(
    delta: float = 1.0,
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return Huber(
        delta=delta,
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.LogCosh",
    name="LogCosh",
)
@controlled_wrapper(LogCosh, wrapper_attribute="__fnwrapped__")
def _LogCosh(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return LogCosh(
        reduction=reduction,
    )


REGRESSION_NODE_SHELFE = Shelf(
    nodes=[
        _MeanSquaredError,
        _MeanAbsoluteError,
        _MeanAbsolutePercentageError,
        _MeanSquaredLogarithmicError,
        _CosineSimilarity,
        _Huber,
        _LogCosh,
    ],
    subshelves=[],
    name="Regression",
    description="",
)


@NodeDecorator(
    node_id="tensorflow.keras.losses.Hinge",
    name="Hinge",
)
@controlled_wrapper(Hinge, wrapper_attribute="__fnwrapped__")
def _Hinge(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return Hinge(
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.SquaredHinge",
    name="SquaredHinge",
)
@controlled_wrapper(SquaredHinge, wrapper_attribute="__fnwrapped__")
def _SquaredHinge(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return SquaredHinge(
        reduction=reduction,
    )


@NodeDecorator(
    node_id="tensorflow.keras.losses.CategoricalHinge",
    name="CategoricalHinge",
)
@controlled_wrapper(CategoricalHinge, wrapper_attribute="__fnwrapped__")
def _CategoricalHinge(
    reduction: Reduction = Reduction.default(),
) -> Loss:
    if isinstance(reduction, Reduction):
        reduction = reduction.value
    return CategoricalHinge(
        reduction=reduction,
    )


HINGE_NODE_SHELFE = Shelf(
    nodes=[_Hinge, _SquaredHinge, _CategoricalHinge],
    subshelves=[],
    name="Hinge",
    description="Hinge losses for 'maximum-margin' classification",
)

LOSSES_NODE_SHELFE = Shelf(
    nodes=[],
    subshelves=[PROBABILISTIC_NODE_SHELFE, REGRESSION_NODE_SHELFE, HINGE_NODE_SHELFE],
    name="Losses",
    description="The purpose of loss functions is to compute the quantity that a model should seek to minimize during training.",
)
